﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;
using System.Windows.Input;
using System.ComponentModel;

namespace SimpleAndFastAvFilterForRealImages
{ public partial class MainPage : UserControl
  { WriteableBitmap img0, img1;
    Int32 kw, hkw;  //kernel width  and half kernel width,  kw uneven and hkw even
    Int32 kh, hkh;  //kernel height and half kernel height, kh uneven and hkh even
    Int32 xSize, ySize; //image width and height
    Int64 t0, t1;       //system time before and after filtering
    String fileName, simple_or_fast = "simple"; 
    Byte [,] R0, G0, B0; //input
    Int32[,] R1, G1, B1; //intermediary
    Int32[,] R2, G2, B2; //output
    BackgroundWorker myThread = new BackgroundWorker(); //Thread object

    public MainPage()
    { InitializeComponent();
      myThread.DoWork += new DoWorkEventHandler( myThreadDoWork    );
      myThread.RunWorkerCompleted += 
             new RunWorkerCompletedEventHandler( myThreadCompleted );
    }

    private void myImageIn_ImageOpened( object sender,RoutedEventArgs e ) 
    { img0  = new WriteableBitmap( (BitmapSource)myImageIn.Source );
      fileName = ((BitmapImage)myImageIn.Source).UriSource.OriginalString;
      a_new_image_opened();
    }

    private void myButtonOF_Click( object sender, RoutedEventArgs e )
    { OpenFileDialog openFileDialog = new OpenFileDialog();
      openFileDialog.Filter = "Image Files (*.png,*.jpg)|*.png;*.jpg";
      bool? userClickedOK = openFileDialog.ShowDialog();
      if ( userClickedOK == true )
      { try
        { System.IO.Stream stream = openFileDialog.File.OpenRead();
          fileName = openFileDialog.File.ToString();
          BitmapImage bmp = new BitmapImage();
          bmp.SetSource( stream );
          stream.Close();
          img0 = new WriteableBitmap( bmp );
          a_new_image_opened();
        } catch {}
      }
    }

    private void a_new_image_opened()
    { xSize = img0.PixelWidth;
      ySize = img0.PixelHeight;
      img1 = new WriteableBitmap( xSize, ySize );
      myImageIn .Source = img0;
      myImageOut.Source = img1;
      //prepare_intermediary_arrays
      R0 = new Byte [ySize, xSize];
      G0 = new Byte [ySize, xSize];
      B0 = new Byte [ySize, xSize];
      R1 = new Int32[ySize, xSize];
      G1 = new Int32[ySize, xSize];
      B1 = new Int32[ySize, xSize];
      R2 = new Int32[ySize, xSize];
      G2 = new Int32[ySize, xSize];
      B2 = new Int32[ySize, xSize];
      GC.Collect(); //clean up memory when new arrays override old ones
      for ( int y=0; y < ySize; y++ )
        for ( int x=0; x < xSize; x++ )
        { Int32 c = img0.Pixels[y*xSize+x];
          R0[y,x] = (Byte)( c & 0x0000ff)       ;
          G0[y,x] = (Byte)((c & 0x00ff00) >>  8);
          B0[y,x] = (Byte)((c & 0xff0000) >> 16);
        }
      kw_Slider .Maximum = xSize;
      kh_Slider.Maximum = ySize;
      if ( xSize > 300 && simple_or_fast == "simple" ) 
        kw_Slider.Value = kh_Slider.Value =  1; 
      else
        kw_Slider.Value = kh_Slider.Value = 11;
      myTextBlockOF.Text = fileName + ": " + xSize.ToString() + 'x' + ySize.ToString();
      myThreadStart();
    }
    private void myThreadStart()
    { if ( myThread.IsBusy ) return;
      kw = (Int32)kw_Slider.Value; if (kw%2 == 0) kw--; kwTextBox.Text = kw.ToString(); 
      kh = (Int32)kh_Slider.Value; if (kh%2 == 0) kh--; khTextBox.Text = kh.ToString();
      if ( simple_or_fast == "simple" && (kw*kh*(xSize-kw)*(ySize-kh)) > 100000000 ) 
      { myTextBlockTime.Text = "Filter rejected because it needs > 5 sec.";
        for ( int i=0; i < xSize*ySize; )
          unchecked { img1.Pixels[i++] = (Int32)0xffffffff; }
        img1.Invalidate(); return;
      }
      hkw = kw / 2;
      hkh = kh / 2;
      myTextBlockTime.Text = "Wait !";
      myBorder.Cursor = Cursors.Wait;
      t0 = DateTime.Now.Ticks; 
      myThread.RunWorkerAsync(); //Calls myThreadDoWork( ... )
    }

    private void myThreadDoWork( Object sender, DoWorkEventArgs e )
    { if ( simple_or_fast == "simple" ) simpleAverageFilter();
      else                              fastAverageFilter();
      fill_borders     ();
    }

    private void myThreadCompleted( Object sender,	RunWorkerCompletedEventArgs e )
    { myBorder.Cursor = Cursors.Arrow;
      t1 = DateTime.Now.Ticks;
      myTextBlockTime.Text = "This filter took " + ((t1-t0)/10000).ToString() + " millisec.";
      img1.Invalidate();
    }

    private void myRB_Click( object sender, RoutedEventArgs e ) 
    { if ( (bool)myRBSimple.IsChecked ) simple_or_fast = "simple"; 
      else                              simple_or_fast = "fast";
      myThreadStart();
    }

    private void Slider_MouseLeftButtonUp( object sender, MouseButtonEventArgs e ) 
    { myThreadStart(); 
    }

    private void thread_DoWork( Object sender, DoWorkEventArgs e )
    { if ( simple_or_fast == "simple" ) simpleAverageFilter(); else fastAverageFilter();
      fill_borders();
    }

    private void simpleAverageFilter()
    { Int32 R, G, B, sumR, sumG, sumB, x, xx, xxx, y, yy, yyy, startOfYRow;
      float Rf, Gf, Bf;
      int kernelArea = kw*kh; //no of pixels inside the kernel
      for ( y=hkh; y < ySize-hkh; y++ ) //==================
      { startOfYRow = y*xSize;
        for ( x=hkw; x < xSize-hkw; x++ ) //===============
        { sumR=sumG=sumB=0;
          for ( yy=-hkh; yy <= hkh; yy++ ) //=============
          { yyy = y + yy;
            for ( xx=-hkw; xx <= hkw; xx++ )//========
            { xxx = x + xx;
              sumR+=R0[yyy,xxx];
              sumG+=G0[yyy,xxx];
              sumB+=B0[yyy,xxx];
            } //====== end for (int xx... ================
          } //======== end for (int yy... ==================
          Rf = sumR/kernelArea; R = Convert.ToInt32(Rf)      ;
          Gf = sumG/kernelArea; G = Convert.ToInt32(Gf) <<  8;
          Bf = sumB/kernelArea; B = Convert.ToInt32(Bf) << 16;
          unchecked { img1.Pixels[startOfYRow+x] = (Int32)0xff000000 | R | G | B; }
        } //============ end for (int x... =====================
      } //============== end for (int y... =======================
    }

    private void fastAverageFilter()
    { Int32 x, x0, x1, x2, x3, y, y0, y1, y2, y3;
      for ( y=0; y < ySize; y++ )         //clear R1,G1, B1, R2, G2, B2
        for ( x=hkw; x < xSize-hkw; x++ ) R1[y,x]=G1[y,x]=B1[y,x]=R2[y,x]=G2[y,x]=B2[y,x] = 0;
      //Horizontal summation into the intermediary integer-matrices R1, G1, B1
      for ( y=0; y < ySize; y++ )
      { //add up the first column = column[hkw]
        for ( x=0; x < kw; x++ ) 
        { R1[y,hkw] += R0[y,x];
          G1[y,hkw] += G0[y,x];
          B1[y,hkw] += B0[y,x];
        }
        //add up all further columns from x2=[hkw+1] to x2=[xSize-hkw-1]
        for ( x0=0, x1=hkw, x2=hkw+1, x3=kw; x3 < xSize; x0++, x1++, x2++, x3++ )
        { R1[y,x2] = R1[y,x1] - R0[y,x0] + R0[y,x3];
          G1[y,x2] = G1[y,x1] - G0[y,x0] + G0[y,x3];
          B1[y,x2] = B1[y,x1] - B0[y,x0] + B0[y,x3];
        }
      }
      //Vertical summation into the output integer-matrices R2, G2, B2
      for ( x=hkw; x < xSize-hkw; x++ ) //vertical sums
      { for ( y=0; y < kh; y++ )        //first row
        { R2[hkh,x] += R1[y,x];
          G2[hkh,x] += G1[y,x];
          B2[hkh,x] += B1[y,x];
        }
        for ( y0=0, y1=hkh, y2=hkh+1, y3=kh; y3 < ySize; y0++, y1++, y2++, y3++ )
        { R2[y2,x] = R2[y1,x] - R1[y0,x] + R1[y3,x];
          G2[y2,x] = G2[y1,x] - G1[y0,x] + G1[y3,x];
          B2[y2,x] = B2[y1,x] - B1[y0,x] + B1[y3,x];
        }
      }
      float divisor = kh * kw; //no of pixels inside the kernel
      int yy, R, G, B;     
      for ( y=hkh; y < ySize-hkh; y++ )   //divide, round, shift and or together
      { yy = y * xSize;
        for ( x=hkw; x < xSize-hkw; x++ ) 
        { R = Convert.ToInt32( (float)R2[y,x] / divisor );
          G = Convert.ToInt32( (float)G2[y,x] / divisor ) <<  8;
          B = Convert.ToInt32( (float)B2[y,x] / divisor ) << 16;
          unchecked { img1.Pixels[yy+x] = (Int32)0xff000000 | B | G | R; }
        }
      }
    }

    private void fill_borders() //copy the outmost valuable pixels to the empty spaces 
    { int x, y, yy, yyy;
      //upper border
      yyy = hkh * xSize;
      for ( y=0; y < hkh; y++ )
      { yy = y * xSize; 
        for ( x=hkw; x < xSize-hkw; x++ )   img1.Pixels[yy+x] = img1.Pixels[yyy+x];
      }
      //lower border
      yyy = (ySize - 1 - hkh) * xSize;
      for ( y=ySize-hkh; y < ySize; y++ )
      { yy = y * xSize; 
        for ( x=hkw; x < xSize-hkw; x++ )   img1.Pixels[yy+x] = img1.Pixels[yyy+x];
      }
      //left border
      for ( y=0; y < ySize; y++ )
      { yy = y * xSize;
        yyy = yy + hkw; 
        for ( x=0; x < hkw; x++ )           img1.Pixels[yy+x] = img1.Pixels[yyy  ];
      }
      //right border
      for ( y=0; y < ySize; y++ )
      { yy = y * xSize;
        yyy = yy + xSize - 1 - hkw;
        for ( x=xSize-hkw; x < xSize; x++ ) img1.Pixels[yy+x] = img1.Pixels[yyy  ];
      }
    }
  }
}